/* -*-Mode: C;-*-
 * $Id: edsiotest.c 1.2 Tue, 13 Jun 2000 01:19:56 -0700 jmacd $
 *
 * Copyright (C) 1998, 1999, 2000, Joshua P. MacDonald
 * <jmacd@CS.Berkeley.EDU> and The Regents of the University of
 * California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 *    Neither name of The University of California nor the names of
 *    its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "edsio.h"
#include "edsiostdio.h"

void
test1 ()
{
  guint32 x = 0;

  PropTest *pt = g_new0 (PropTest, 1);

  EdsioPropTestUintProperty prop;

  g_assert (edsio_library_init ());

  g_assert (edsio_new_proptest_uint_property ("Just testing (1)...", PF_Persistent, & prop));

  g_assert (! proptest_test_uint (pt, prop));

  g_assert (proptest_set_uint (pt, prop, 42));

  g_assert (proptest_test_uint (pt, prop));

  g_assert (proptest_get_uint (pt, prop, & x) && x == 42);

  /* kill the cache, to test persistence. */
  pt->_edsio_property_table = NULL;

  g_assert (proptest_get_uint (pt, prop, & x) && x == 42);

  g_assert (proptest_unset_uint (pt, prop));

  g_assert (! proptest_test_uint (pt, prop));

  g_assert (! pt->_edsio_property_table);
}

void
test2 ()
{
  const char* str = "hello there";
  const char* str2;
  guint32 str2_len;

  PropTest *pt = g_new0 (PropTest, 1);

  EdsioPropTestBytesProperty prop;

  g_assert (edsio_edsio_init ());

  g_assert (edsio_new_proptest_bytes_property ("Just testing (2)...", PF_Persistent, & prop));

  g_assert (! proptest_test_bytes (pt, prop));

  g_assert (proptest_set_bytes (pt, prop, (guint8*) str, strlen (str) + 1));

  g_assert (proptest_test_bytes (pt, prop));

  g_assert (proptest_get_bytes (pt, prop, (const guint8**) & str2, & str2_len) && str2_len == (strlen (str) + 1) && strcmp (str, str2) == 0);

  /* kill the cache, to test persistence. */
  pt->_edsio_property_table = NULL;

  g_assert (proptest_get_bytes (pt, prop, (const guint8**) & str2, & str2_len) && str2_len == (strlen (str) + 1) && strcmp (str, str2) == 0);

  g_assert (proptest_unset_bytes (pt, prop));

  g_assert (! proptest_test_bytes (pt, prop));

  g_assert (! pt->_edsio_property_table);
}

void
test3 ()
{
  SerialEdsioUint x;
  SerialEdsioUint *p;
  EdsioPropTestEdsioUintProperty prop;

  PropTest *pt = g_new0 (PropTest, 1);

  x.val = 42;

  g_assert (edsio_edsio_init ());

  g_assert (edsio_new_proptest_edsiouint_property ("Just testing (3)...", PF_Persistent, & prop));

  g_assert (! proptest_test_edsiouint (pt, prop));

  g_assert (proptest_set_edsiouint (pt, prop, & x));

  g_assert (proptest_test_edsiouint (pt, prop));

  g_assert (proptest_get_edsiouint (pt, prop, & p) && x.val == p->val);

  /* kill the cache, to test persistence. */
  pt->_edsio_property_table = NULL;

  g_assert (proptest_get_edsiouint (pt, prop, & p) && x.val == p->val);

  g_assert (proptest_unset_edsiouint (pt, prop));

  g_assert (! proptest_test_edsiouint (pt, prop));

  g_assert (! pt->_edsio_property_table);
}

void
test4 ()
{
  const char* str = "hello there";
  const char* str2;

  PropTest *pt = g_new0 (PropTest, 1);

  EdsioPropTestStringProperty prop;

  g_assert (edsio_edsio_init ());

  g_assert (edsio_new_proptest_string_property ("Just testing (4)...", PF_Persistent, & prop));

  g_assert (! proptest_test_string (pt, prop));

  g_assert (proptest_set_string (pt, prop, str));

  g_assert (proptest_test_string (pt, prop));

  g_assert (proptest_get_string (pt, prop, & str2) && strcmp (str, str2) == 0);

  /* kill the cache, to test persistence. */
  pt->_edsio_property_table = NULL;

  g_assert (proptest_get_string (pt, prop, & str2) && strcmp (str, str2) == 0);

  g_assert (proptest_unset_string (pt, prop));

  g_assert (! proptest_test_string (pt, prop));

  g_assert (! pt->_edsio_property_table);
}

void test5 ()
{
  GByteArray* sink_result;
  SerialSink* sink = edsio_simple_sink (NULL, SBF_Checksum | SBF_Base64 | SBF_Compress, FALSE, NULL, & sink_result);
  SerialSource* src;
  const char* input = "hello world";
  SerialEdsioString *output;
  guint8 zero[1];
  zero[0] = 0;

  g_assert (serialize_edsiostring (sink, input));

  g_assert (sink->sink_close (sink));

  g_byte_array_append (sink_result, zero, 1);

  handle_printf (_stderr_handle, "%s -> %s\n", input, sink_result->data);

  src = edsio_simple_source (sink_result->data, sink_result->len - 1, SBF_Checksum | SBF_Base64 | SBF_Compress);

  g_assert (unserialize_edsiostring (src, & output));

  g_assert (src->source_close (src));

  g_assert (strcmp (output->val, input) == 0);
}

void
test6 ()
{
  const char* md5str = "aed3918c4ccb89f2dcf74d5dcab22989";
  const char* md5strbad1 = "aed3918c4cXb89f2dcf74d5dcab22989";
  const char* md5strbad2 = "aed3918c4cab89f2dcf74d5dcab22989X";
  const char* md5strbad3 = "aed3918c4cab89f2dcf74d5dcab2298";
  char md5str2[33];
  guint8 md5[16];

  g_assert (! edsio_md5_from_string (md5, md5strbad1));
  g_assert (! edsio_md5_from_string (md5, md5strbad2));
  g_assert (! edsio_md5_from_string (md5, md5strbad3));

  g_assert (edsio_md5_from_string (md5, md5str));

  edsio_md5_to_string (md5, md5str2);

  g_assert (strcmp (md5str, md5str2) == 0);
}

#if 0
void
test7 ()
{
  guint8 buf[1024];
  gint n;
  const guint8* in, *out;
  const MessageDigest* md = edsio_message_digest_md5 ();
  char mdbuf[33];

  g_assert (handle_digest_compute (_stdin_handle, md));
  g_assert (handle_digest_compute (_stdout_handle, md));

  for (;;)
    {
      n = handle_read (_stdin_handle, buf, 1024);

      g_assert (n >= 0);

      g_assert (handle_write (_stdout_handle, buf, n));

      if (n < 1024)
	break;
    }

  g_assert (handle_close (_stdin_handle));
  g_assert (handle_close (_stdout_handle));

  g_assert (in = handle_digest (_stdin_handle, md));
  g_assert (out = handle_digest (_stdout_handle, md));

  edsio_md5_to_string (in, mdbuf);
  handle_printf (_stderr_handle, "in: %s\n", mdbuf);

  edsio_md5_to_string (out, mdbuf);
  handle_printf (_stderr_handle, "out: %s\n", mdbuf);
}
#endif

void
test8 ()
{
  SerialTest st;
  guint32 b[4];
  SerialGenericTime d[1];

  b[0] = 1;
  b[1] = 3;
  b[2] = 5;
  b[3] = 7;

  st.b_len = 4;
  st.b = b;

  st.c_len = 0;

  d[0].seconds = 53;
  d[0].nanos = 55;
  st.d_len = 1;
  st.d = d;

  serializeio_generic_print (ST_Test, & st, 2);
}

void test9 ()
{
  const char* files[] = {
    "/dev/null",
    "edsiotest.o",
    "../LICENSE",
    "Makefile.in"
  };
  int i;

  for (i = 0; i < 4; i += 1)
    {
      const char* file = files[i];

      FileHandle *fsrfh;
      FileHandle *mwfh;
      FileHandle *mrfh;
      const guint8 *buf;
      gint diff;
      guint len;

      if (! (fsrfh = handle_read_file (file)) ||
	  ! (mwfh =  handle_write_mem ()))
	abort ();

      if (! handle_drain (fsrfh, mwfh))
	abort ();

      if (! handle_close (mwfh) ||
	  ! handle_close (fsrfh))
	abort ();

      handle_free (fsrfh);

      if (! (fsrfh = handle_read_file (file)))
	abort ();

      handle_write_mem_result (mwfh, & buf, & len);

      if (! (mrfh = handle_read_mem (buf, len)))
	abort ();

      if (! handle_cmp (mrfh, fsrfh, NULL, & diff))
	abort ();

      g_assert (! diff);

      handle_close (mrfh);
      handle_close (fsrfh);

      handle_free (mrfh);
      handle_free (fsrfh);
      handle_free (mwfh);
    }
}

int
main ()
{
  test1 ();

  test2 ();

  test3 ();

  test4 ();

  test5 ();

  test6 ();

#if 0
  test7 ();
#endif

  test8 ();

  test9 ();

  return 0;
}
