/* -*-Mode: C;-*-
 * $Id: rcsextract.c 1.6 Wed, 14 Jun 2000 22:01:07 -0700 jmacd $
 *
 * Copyright (C) 2000, Joshua P. MacDonald <jmacd@CS.Berkeley.EDU>
 * and The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 *    Neither name of The University of California nor the names of
 *    its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <edsio.h>
#include <edsiostdio.h>
#include <ctype.h>
#include "xdfs.h"

/* Warning: very cheesy!
 */

#ifdef DEBUG_EXTRACT
  FileHandle *fh2 = handle_read_file (filename);

  guint8* debug_buf = g_malloc (buflen);

  if (! handle_read (fh2, debug_buf, buflen))
    g_error ("read failed");
#endif

gboolean
rcs_count (const char* filename, guint *encoded_size)
{
  char *readbuf0, *readbuf;
  gboolean in_string = FALSE;
  gboolean in_text = FALSE;
  guint string_start = 0;
  guint string_end = 0;
  guint current_pos = 0;
  /*char *current_delta = NULL;*/
  FileHandle *fh = handle_read_file (filename);
  guint buflen = handle_length (fh);

  (* encoded_size) = 0;

  readbuf0 = g_new (guint8, buflen);

  for (;;)
    {
      int c = handle_gets (fh, readbuf0, buflen);

      readbuf = readbuf0;

      if (c < 0)
	break;

      if (strncmp (readbuf, "text", 4) == 0)
	in_text = TRUE;

      if (! in_string && readbuf[0] == '@')
	{
	  string_start = current_pos + 1;
	  in_string = TRUE;
	  readbuf += 1;
	}

      current_pos += c;

      if (in_string)
	{
	  while ((readbuf = strchr (readbuf, '@')))
	    {
	      if (readbuf[1] == '@')
		{
		  string_start += 1; /* @@@ bogus, just counting. */
		  readbuf += 2;
		  continue;
		}

	      in_string = FALSE;
	      break;
	    }

	  string_end = current_pos - 2;

	  if (in_text && ! in_string)
	    {
	      in_text = FALSE;

	      /*g_free (current_delta);
		current_delta = NULL;*/

	      (* encoded_size) += (string_end - string_start);
	    }

	  continue;
	}

      if (isdigit (readbuf[0]))
	{
#if 0
	  (* strchr (readbuf, '\n')) = 0;
	  if (current_delta)
	    g_free (current_delta);
	  current_delta = g_strdup (readbuf);
#endif
	}
    }

  handle_close (fh);

  g_free (readbuf0);

#if 0
  if (current_delta)
    g_free (current_delta);
#endif

  return TRUE;
}

#if 0
int
main (int argc, char** argv)
{
  guint size;

  if (argc != 2)
    g_error ("usage: %s RCS_file\n", argv[0]);

  if (! rcs_count (argv[1], &size))
    g_error ("rcs_parse failed");

  return 0;
}
#endif
